class TabsCache {
	constructor() {
		//store and manage info about tabs
		this.altUrls = {}; // cache_alt_urls
		this.serverRedirects = {}; // cache_redirects
		this.clientRedirects = {}; // cache_credirects
		this.sublasts = {}; // sublasts
		this.headers = {}; // headers
		this.createdTimestamp = {}; // cache_created_tabs_ts
		this.hashTabs = {}; // hash_tabs
		this.lastUrls = {}; // last_urls
		this.prevTabId = {}; // prev_tab_id
	}
	// clean up
	cleanUp(tabId) {
		if (!tabId) {
			return;
		}
		this.deleteServerRedirect(tabId);
		this.deleteClientRedirect(tabId);
		this.removeSublast(tabId);
		this.removePrevId(tabId);
	}
	// headers
	setHeadersFlag(tabId, value) {
		this.headers[tabId] ? this.headers[tabId].flag = value : this.headers[tabId] = { flag: value, referer: "" };
	}
	setRefererHeader(tabId, referer) {
		this.headers[tabId] ? this.headers[tabId].referer = referer : this.headers[tabId] = { flag: false, referer };
	}
	removeHeader(tabId) {
		delete this.headers[tabId];
	}
	hasHeadersFlag(tabId, referer) {
		return this.headers[tabId] ? this.headers[tabId].flag : false;
	}
	getRefererHeader(tabId) {
		return this.headers[tabId] ? this.headers[tabId].referer : "";
	}
	resetRefererForDuplicatesAndReopened(tabId) {
		if (!this.hashTabs[tabId]) {
			return;
		}
		if (this.hashTabs[tabId].isDuplicate || this.hashTabs[tabId].isReopened) {
			this.setRefererHeader(tabId, "");
			this.hashTabs[tabId].isDuplicate = false;
			this.hashTabs[tabId].isReopened = false;
		}
	}
	// sublasts
	getSublast(tabId) {
		return this.isSublastExist(tabId) ? this.sublasts[tabId] : "";
	}
	addSublast(tabId, fetching, url) {
		this.sublasts[tabId] = {
			fetching: fetching,
			url: url
		};
	}
	setSublastFetching(tabId, fetching) {
		let sublast = this.getSublast(tabId);
		if (sublast && fetching !== undefined) {
			sublast.fetching = fetching;
		}
	}
	isSublastExist(tabId) {
		return this.sublasts.hasOwnProperty(tabId);
	}
	isSublastFetching(tabId) {
		return this.isSublastExist(tabId) && this.getSublast(tabId).fetching;
	}
	removeSublast(tabId) {
		if (this.isSublastExist(tabId)) {
			delete this.sublasts[tabId];
		}
	}
	replaceSublastId(oldId, newId) {
		if (!this.isSublastExist(oldId)) {
			return;
		}
		let oldSublast = this.getSublast(oldId);
		this.addSublast(newId, oldSublast.fetching, oldSublast.url);
		this.removeSublast(oldId);
	}
	// previd
	getPrevId(tabId) {
		return this.isPrevIdExist(tabId) ? this.prevTabId[tabId] : null;
	}
	setPrevId(tabId, prevTabId) {
		this.prevTabId[tabId] = prevTabId;
	}
	isPrevIdExist(tabId) {
		return this.prevTabId.hasOwnProperty(tabId);
	}
	removePrevId(tabId) {
		if (this.isPrevIdExist(tabId)) {
			delete this.prevTabId[tabId];
		}
	}
	// createdTimestamp
	addCreatedTabTimestamp(tabId) {
		if (this.isCreatedTabTimestampExist(tabId)) {
			return;
		}
		this.createdTimestamp[tabId] = Date.now();
	}
	removeCreatedTabTimestamp(tabId) {
		delete this.createdTimestamp[tabId];
	}
	isCreatedTabTimestampExist(tabId) {
		return this.createdTimestamp.hasOwnProperty(tabId);
	}
	isNewWindowTabCreated(tabId) {
		if (!this.isCreatedTabTimestampExist(tabId)) {
			return false;
		}
		return Date.now() - this.createdTimestamp[tabId] < 3000;
	}
	// server redirects
	addServerRedirectInfo(tabId) {
		if (tabId == undefined) {
			throw new Error("addServerRedirectInfo, tabId == undefined");
		}
		if (this.serverRedirects.hasOwnProperty(tabId) == false) {
			this.serverRedirects[tabId] = new ServerRedirectsInfo();
			return this.serverRedirects[tabId];
		} else {
			throw new Error("addServerRedirectInfo, already exists, tabId = " + tabId);
		}
	}
	getServerRedirectInfo(tabId) {
		if (tabId == undefined) {
			throw new Error("getServerRedirectInfo, tabId == undefined");
		}
		if (this.serverRedirects.hasOwnProperty(tabId) == false) {
			return null;
		} else {
			return this.serverRedirects[tabId];
		}
	}
	isServerRedirectExist(tabId) {
		return this.getServerRedirectInfo() ? true : false;
	}
	getServerRedirects(tabId) {
		if (tabId == undefined) {
			throw new Error("getServerRedirects, tabId == undefined");
		}
		if (this.serverRedirects.hasOwnProperty(tabId) == false) {
			return null;
		} else {
			if (this.serverRedirects[tabId].redirectUrls) {
				return this.serverRedirects[tabId].redirectUrls;
			} else {
				return null;
			}
		}
	}
	deleteServerRedirect(tabId, transitionType) {
		if (tabId == undefined) {
			throw new Error("deleteServerRedirect, tabId == undefined");
		}
		if (this.serverRedirects.hasOwnProperty(tabId) == false) {
			return null;
		} else {
			return delete this.serverRedirects[tabId];
		}
	}
	// client redirects
	addClientRedirect(tabId, url) {
		if (tabId == undefined) {
			throw new Error("addClientRedirect, tabId == undefined");
		}
		this.clientRedirects[tabId] = url;
	}
	deleteClientRedirect(tabId, url) {
		if (tabId == undefined) {
			throw new Error("deleteClientRedirect, tabId == undefined");
		}
		delete this.clientRedirects[tabId];
	}
	getClientRedirect(tabId) {
		if (tabId == undefined) {
			throw new Error("getClientRedirect, tabId == undefined");
		}
		if (this.isClientRedirectExist(tabId)) {
			return this.clientRedirects[tabId];
		}
		return null;
	}
	isClientRedirectExist(tabId) {
		return this.clientRedirects.hasOwnProperty(tabId);
	}
}
class ServerRedirectsInfo {
	constructor() {
		this.redirectUrls = [];
		this.completedUrl = "";
	}
	addUrl(url) {
		this.redirectUrls.push(url);
	}
	// getters / setters
	setCompletedUrl(url) {
		this.completedUrl = url;
	}
	getCompletedUrl() {
		return this.completedUrl;
	}
	getRedirectUrls() {
		return this.redirectUrls;
	}
}
class TabInfo {
	constructor(tab) {
		this.tab = tab;
		this.requestsIds = [];
		this.isDuplicate = false;
		this.isReopened = false;
	}
	addRequestId(requestId) {
		this.requestsIds.push(requestId);
	}
	addCommittedNavigation(navigationInfo) {
		this.committedNavigations.push(navigationInfo);
	}
	isRequestIdExist(requestId) {
		return this.requestsIds.indexOf(requestId) !== -1;
	}
	getTab() {
		return this.tab;
	}
	getRequestIds() {
		return this.requestsIds;
	}
}